import keyword
import re
import string
from collections import namedtuple
from unicodedata import combining
from unicodedata import normalize
from urllib.parse import urlparse

import stringcase

VarMakerStrategy = namedtuple("VarMakerStrategy", "case_func,special_replacer")

_case_map = {
    "snake": VarMakerStrategy(stringcase.snakecase, "_"),
    "pascal": VarMakerStrategy(stringcase.pascalcase, ""),
    "camel": VarMakerStrategy(stringcase.camelcase, ""),  # :(
}

_number_map = {
    "drop": lambda v: v.lstrip(string.digits),
    "underscore": lambda v: f"_{v}",
}


def make_var_name(s: str, case="snake", number_strategy="drop"):
    case_func, special_replacer = _case_map[case]

    # replace special chars, and strip them from the ends
    v = re.sub(r"[\W_]+", special_replacer, s)
    v = v.strip(special_replacer)

    # force consecutive caps to not split things up
    minimizing = False
    s = ""
    for c in v:
        if c.isupper() and minimizing:
            c = c.lower()
        else:
            minimizing = c.isupper()
        s += c

    # do the actual conversion and handle leading numbers
    v = case_func(s)
    v = _number_map[number_strategy](v) if v[0].isnumeric() else v

    if v[0] == "_" and number_strategy != "underscore":
        v = v.lstrip("_")
    v = v.replace("__", "_")  # patch for a_b_c => a__b__c

    return v + ("_" if keyword.iskeyword(v) else "")


def make_unique_var_name(s: str, taken, case="snake", number_strategy="drop"):
    var_name = make_var_name(s, case, number_strategy)
    if var_name in taken:
        for i in range(2, 1000):
            if f"{var_name}_{i}" not in taken:
                return f"{var_name}_{i}"
    return var_name


def id_from_iso_string(s):
    return (
        s.replace(":", "")
        .replace("T", "")
        .replace("Z", "")
        .replace("-", "")
        .replace(".", "")
    )


def parse_realm_and_app_id_from_url(url):
    url = f"https://{url}" if "https://" not in url else url
    parsed = urlparse(url)
    return parsed.hostname, parsed.path.split("/")[-1]


def normalize_hostname(realm_hostname: str) -> str:
    return realm_hostname.lstrip("https://").lstrip("http://").rstrip("/")


def normalize_unicode(s: str) -> str:
    # from: https://stackoverflow.com/a/517974/612166
    nfkd_form = normalize("NFKD", s)
    return "".join([c for c in nfkd_form if not combining(c)])



import json

import requests

from quickbase_client.orm.serialize import QuickbaseJsonEncoder


class QuickbaseRequestFactory(object):
    def __init__(
        self,
        user_token,
        realm_hostname,
        agent="python",
        encoder=None,
        allow_deletes=False,
    ):
        self.user_token = user_token
        self.realm_hostname = realm_hostname
        self.agent = agent
        self.encoder = QuickbaseJsonEncoder() if encoder is None else encoder
        self.allow_deletes = allow_deletes
        self.session = requests.Session()

    @property
    def base_headers(self):
        return {
            "Content-Type": "application/json",
            "User-Agent": self.agent,
            "Authorization": f"QB-USER-TOKEN {self.user_token}",
            "QB-Realm-Hostname": self.realm_hostname,
        }

    def make_request(
        self, method, endpoint, additional_headers=None, params=None, data=None
    ) -> requests.Response:
        """Make a request (synchronously) and return the Response.

        :param string method: The (string) HTTP method.
        :param string endpoint: The endpoint of the API (starting after "v1/" for example).
        :param dict additional_headers: A dict of extra headers.
        :param params: Query parameters.
        :param data: The data to send in the body.
        """
        additional_headers = {} if additional_headers is None else additional_headers
        headers = {**self.base_headers, **additional_headers}
        url = f'https://api.quickbase.com/v1/{endpoint.lstrip("/")}'
        if method == "DELETE" and not self.allow_deletes:
            raise RuntimeError("to allow deletes, please set allow_deletes=True")
        # ew - but easiest way to serialize dates
        data = json.loads(self.encoder.encode(data))
        return self.session.request(
            method=method.upper(), url=url, headers=headers, json=data, params=params
        )

    def get(self, endpoint, additional_headers=None, params=None):
        return self.make_request("GET", endpoint, additional_headers, params=params)

    def post(self, endpoint, additional_headers=None, params=None, data=None):
        return self.make_request(
            "POST", endpoint, additional_headers, params=params, data=data
        )

    def delete(self, endpoint, additional_headers=None, data=None):
        return self.make_request("DELETE", endpoint, additional_headers, data=data)


QuickBaseRequestFactory = QuickbaseRequestFactory  # alias - TODO - delete in future



from enum import Enum

import attr


class QuickbaseFieldType(Enum):
    """An Enumeration of Field Types.

    .. note::
        Formula fields use the underlying type.

    In the generated classes, the import for this class is usually aliased
    as ``Qb`` to make it easier to write like ``Qb.TEXT``.

    These also all have constants in ``quickbase_client.orm.field`` module prefixed with
    ``QB_``.
    """

    TEXT = 100
    TEXT_MULTILINE = 101
    TEXT_MULTIPLE_CHOICE = 102
    TEXT_MULTI_SELECT = 102
    RICH_TEXT = 103
    NUMERIC = 200
    NUMERIC_CURRENCY = 201
    NUMERIC_PERCENT = 202
    NUMERIC_RATING = 203
    DATE = 300
    DATETIME = 301
    TIME_OF_DAY = 302
    DURATION = 303
    CHECKBOX = 400
    ADDRESS = 500
    EMAIL_ADDRESS = 501
    USER = 600
    OTHER = 900


QuickBaseFieldType = QuickbaseFieldType  # alias - TODO - delete in future


QB_TEXT = QuickbaseFieldType.TEXT
QB_TEXT_MULTILINE = QuickbaseFieldType.TEXT_MULTILINE
QB_TEXT_MULTIPLE_CHOICE = QuickbaseFieldType.TEXT_MULTIPLE_CHOICE
QB_TEXT_MULTI_SELECT = QuickbaseFieldType.TEXT_MULTI_SELECT
QB_RICH_TEXT = QuickbaseFieldType.RICH_TEXT
QB_NUMERIC = QuickbaseFieldType.NUMERIC
QB_NUMERIC_CURRENCY = QuickbaseFieldType.NUMERIC_CURRENCY
QB_NUMERIC_PERCENT = QuickbaseFieldType.NUMERIC_PERCENT
QB_NUMERIC_RATING = QuickbaseFieldType.NUMERIC_RATING
QB_DATE = QuickbaseFieldType.DATE
QB_DATETIME = QuickbaseFieldType.DATETIME
QB_TIME_OF_DAY = QuickbaseFieldType.TIME_OF_DAY
QB_DURATION = QuickbaseFieldType.DURATION
QB_CHECKBOX = QuickbaseFieldType.CHECKBOX
QB_ADDRESS = QuickbaseFieldType.ADDRESS
QB_EMAIL_ADDRESS = QuickbaseFieldType.EMAIL_ADDRESS
QB_USER = QuickbaseFieldType.USER
QB_OTHER = QuickbaseFieldType.OTHER


_qb_api_type_lookup = {
    "text": QuickbaseFieldType.TEXT,
    "text-multi-line": QuickbaseFieldType.TEXT_MULTILINE,
    "multitext": QuickbaseFieldType.TEXT_MULTI_SELECT,
    "checkbox": QuickbaseFieldType.CHECKBOX,
    "timestamp": QuickbaseFieldType.DATETIME,
    "numeric": QuickbaseFieldType.NUMERIC,
    "currency": QuickbaseFieldType.NUMERIC_CURRENCY,
    "percent": QuickbaseFieldType.NUMERIC_PERCENT,
    "rating": QuickbaseFieldType.NUMERIC_RATING,
    "duration": QuickbaseFieldType.DURATION,
    "date": QuickbaseFieldType.DATE,
    "timeofday": QuickbaseFieldType.TIME_OF_DAY,
    "rich-text": QuickbaseFieldType.RICH_TEXT,
    "text-multiple-choice": QuickbaseFieldType.TEXT_MULTIPLE_CHOICE,
}


def get_field_type_by_string(s):
    return _qb_api_type_lookup.get(s, QuickbaseFieldType.OTHER)


@attr.s(auto_attribs=True)
class QuickbaseField(object):
    """The metadata for a specific field.

    :var int fid: The field id.
    :ivar QuickbaseFieldType field_type: The :class:`~QuickbaseFieldType` of the field.
    :ivar str label: The label for the field.
    :ivar Optional[str] formula: The Quickbase string formula.
    """

    fid: int
    field_type: QuickbaseFieldType
    label: str = ""
    formula: str = None

    @property
    def is_formula(self):
        return self.formula is not None and self.formula != ""


# FUTURE - could actually store more type-specific field metadata

QuickBaseField = QuickbaseField  # alias - TODO - delete in future



import requests

from quickbase_client.utils.string_utils import normalize_hostname


class QuickbaseLegacyApiClient:
    """Legacy Client which makes requests to the old XML API.

    This operates more generically to send any request to the old API. It's primary
    purpose is to supplement the JSON API for things not yet supported.

    It does include some higher-level methods for things that in particular are
    not supported in the new API's. If there are others you would like added,
    submit an Issue or a Merge-Request and we can certainly add it!

    Rather than get in to having lxml as an optional dependency, it will just deal
    with XML data as strings for now. If we want more support of the XML API, then
    it might be worth it to introduce that.

    :param str user_token:
        The user token for authenticating with the API. Note that the other forms of
        authentication through the XML API are not supported via this library.


    :param str realm_hostname:
        The hostname - like ``"foo.quickbase.com"``

    """

    def __init__(self, user_token: str, realm_hostname: str):
        self.user_token = user_token
        self.realm_hostname = normalize_hostname(realm_hostname)
        self.session = requests.Session()

    def make_request(
        self,
        http_method: str,
        quickbase_action: str,
        endpoint: str,
        request_data_xml_str="",
    ):
        """Used as a simple Python interface to the old XML API.

        :param str http_method:
            The HTTP method, like ``"post"``

        :param str quickbase_action:
            The Quickbase Action to send, like ``"API_ChangeRecordOwner"``

        :param str endpoint:
            The HTTP Endpoint to call, like ``/db/abc123``

        :param str request_data_xml_str:
            The data to add to the XML string that gets sent, *inside* of the
            ``qdbapi`` element/tag. For now this is all done through strings, in the
            future it may use lxml and have better capabilities for working with
            those trees.
        """

        data_xml = f"""
        <qdbapi>
           <usertoken>{self.user_token}</usertoken>
           {request_data_xml_str}
        </qdbapi>
        """
        headers = {
            "Content-Type": "application/xml",
            "QUICKBASE-ACTION": quickbase_action,
        }
        url = f"https://{self.realm_hostname}/{endpoint.lstrip('/')}"
        self.session.request(
            method=http_method,
            url=url,
            headers=headers,
            data=data_xml,
        )

    # Common Use Cases:

    def change_record_owner(self, table_id, rid, new_owner):
        """The Quickbase API_ChangeRecordOwner action

        See https://help.quickbase.com/api-guide/change_record_owner.html

        :param table_id:
            The table ID containing the record

        :param rid:
            The record ID to change the owner of

        :param new_owner:
            The email address, or user ID, of the user to change to the owner.
        """
        return self.make_request(
            http_method="POST",
            quickbase_action="API_ChangeRecordOwner",
            endpoint=f"/db/{table_id}",
            request_data_xml_str=f"""
                <rid>{rid}</rid>
                <newowner>{new_owner}</newowner>
            """,
        )



from quickbase_client.client.legacy_client import QuickbaseLegacyApiClient
from quickbase_client.client.request_factory import QuickbaseRequestFactory


def make_payload(d):
    return {k: v for k, v in d.items() if v is not None}


class QuickbaseApiClient(object):
    """The lower-level client to make API requests.

    .. note::
        Current alias of ``QuickBaseApiClient`` for backwards compatibility - will be
        removed in version 1.0

    Use :meth:`~request` to make an arbitrary request that forwards to
    :meth:`~QuickbaseRequestFactory.make_request`

    :ivar legacy_api:
        The :class:`~QuickbaseLegacyApiClient` for making requests to the XML API.
    """

    def __init__(self, user_token, realm_hostname, agent="python", allow_deletes=False):
        self.rf = QuickbaseRequestFactory(
            user_token, realm_hostname, agent, allow_deletes=allow_deletes
        )
        self.legacy_api = QuickbaseLegacyApiClient(user_token, realm_hostname)

    def request(self, *args, **kwargs):
        return self.rf.make_request(*args, **kwargs)

    # FUTURE - Can add more of these methods, even refactoring in to separate classes
    #   for each main endpoint. Initially just including some of the more common
    #   functions.

    def get_app(self, app_id):
        return self.rf.get(f"/apps/{app_id}")

    def get_tables_for_app(self, app_id):
        return self.rf.get("/tables", params={"appId": app_id})

    def get_table(self, app_id, table_id):
        return self.rf.get(f"/tables/{table_id}", params={"appId": app_id})

    def get_fields_for_table(self, table_id):
        return self.rf.get("/fields", params={"tableId": table_id})

    def get_field(self, field_id, table_id):
        return self.rf.get(f"/fields/{field_id}", params={"tableId": table_id})

    def get_reports_for_table(self, table_id):
        return self.rf.get("/reports", params={"tableId": table_id})

    def get_report(self, report_id, table_id):
        return self.rf.get(f"/reports/{report_id}", params={"tableId": table_id})

    def run_report(self, report_id, table_id, skip=None, top=None):
        payload = make_payload({"skip": skip, "top": top})
        return self.rf.post(
            f"/reports/{report_id}/run", params={"tableId": table_id}, data=payload
        )

    def add_records(
        self, table_id, data=None, merge_field_id=None, fields_to_return=None
    ):
        return self.rf.post(
            "/records",
            data=make_payload(
                {
                    "to": table_id,
                    "data": data,
                    "mergeFieldId": merge_field_id,
                    "fieldsToReturn": fields_to_return,
                }
            ),
        )

    def query(
        self,
        table_id,
        fields_to_select=None,
        where_str=None,
        sort_by=None,
        group_by=None,
        options=None,
    ):
        return self.rf.post(
            "/records/query",
            data=make_payload(
                {
                    "from": table_id,
                    "select": fields_to_select,
                    "where": where_str,
                    "sortBy": sort_by,
                    "groupBy": group_by,
                    "options": options,
                }
            ),
        )


QuickBaseApiClient = QuickbaseApiClient  # alias - TODO - delete in future



import abc


class Script(abc.ABC):
    registration_name = None

    @abc.abstractmethod
    def run(self):
        pass

    @staticmethod
    @abc.abstractmethod
    def add_argparse_args(parser):
        pass

    @staticmethod
    @abc.abstractmethod
    def instantiate_from_ns(ns) -> "Script":
        pass



from quickbase_client.query.ast import after_
from quickbase_client.query.ast import and_
from quickbase_client.query.ast import before_
from quickbase_client.query.ast import contains_
from quickbase_client.query.ast import during_
from quickbase_client.query.ast import eq_
from quickbase_client.query.ast import gt_
from quickbase_client.query.ast import gte_
from quickbase_client.query.ast import has_
from quickbase_client.query.ast import lt_
from quickbase_client.query.ast import lte_
from quickbase_client.query.ast import not_contains_
from quickbase_client.query.ast import not_during_
from quickbase_client.query.ast import not_eq_
from quickbase_client.query.ast import not_has_
from quickbase_client.query.ast import not_starts_width_
from quickbase_client.query.ast import on_or_after_
from quickbase_client.query.ast import on_or_before_
from quickbase_client.query.ast import or_
from quickbase_client.query.ast import starts_with_
from quickbase_client.query.ast import true_


__all__ = [
    "or_",
    "and_",
    "contains_",
    "not_contains_",
    "has_",
    "not_has_",
    "eq_",
    "true_",
    "not_eq_",
    "starts_with_",
    "not_starts_width_",
    "before_",
    "on_or_before_",
    "after_",
    "on_or_after_",
    "during_",
    "not_during_",
    "lt_",
    "lte_",
    "gt_",
    "gte_",
]



class QuickbaseQuery(object):
    """A base object for all of the data for a query.

    .. note::
        Current alias of ``QuickbaseQuery`` for backwards compatibility.

    :param str where: The where string, e.g. ``"{7.EX.'18'}"``

    :param dict options:
        Additional options to pass to the Quickbase runQuery endpoint.

    :param list[dict] group_by:
        The groupBy for the Quickbase runQuery endpoint.

    :param list[dict] sort_by:
        The sortBy for the Quickbase runQuery endpoint.

    :param list[int] select:
        The list of field ID's to return. Note that Quickbase by default (i.e. if
        this parameter is left as ``None``) only returns the "default" fields for
        the table.
    """

    def __init__(self, where, options=None, group_by=None, sort_by=None, select=None):
        self.where = where
        self.options = options
        self.group_by = group_by
        self.sort_by = sort_by
        self.select = select


QuickBaseQuery = QuickbaseQuery



try:
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import version

    try:
        __version__ = version("quickbase_client")
    except PackageNotFoundError:
        __version__ = None
except ModuleNotFoundError:
    __version__ = None

if __name__ == "__main__":
    print(__version__)



class ResponsePager:
    """Object to pass to methods (query) to manage pagination.

    When calling something like QuickbaseTableClient.query, you can pass a ResponsePager,
    and repeatedly make requests while :meth:`~more_remaining` is True.

    .. code:: python

       pager = ResponsePager()
       while pager.more_remaining():
           recs = my_client.query(pager=pager)

    """

    def __init__(self):
        self.num_calls = 0
        self.next_skip = 0
        self.total_records = -1

    def update_from_metadata(self, metadata):
        self.num_calls += 1
        self.total_records = metadata["totalRecords"]
        self.next_skip = metadata["skip"] + metadata["numRecords"]

    def get_options(self):
        return {"skip": self.next_skip}

    def more_remaining(self) -> bool:
        """Returns true if there is another request to be made."""
        return self.total_records == -1 or self.next_skip < self.total_records



from typing import Any
from typing import List
from typing import Type
from typing import Union

from quickbase_client import ResponsePager
from quickbase_client.client.api import QuickbaseApiClient
from quickbase_client.orm.field import QuickbaseField
from quickbase_client.orm.report import QuickbaseReport
from quickbase_client.orm.serialize import RecordJsonSerializer
from quickbase_client.orm.table import QuickbaseTable
from quickbase_client.query.query_base import QuickbaseQuery


class QuickbaseTableClient(object):
    """Class for making API calls relative to a specific QuickBase table.

    This includes making calls for the app in general.

    All calls (except :py:meth:`~query`) return a :class:`~requests.Response` object
    for the HTTP response.

    .. note::

        Pagination is not handled in any of these methods (yet).

    :ivar table:
        The underlying :class:`~QuickbaseTable`

    :ivar api:
        The wrapped :class:`~QuickbaseApiClient`

    :param user_token:
        The user token to authenticate.

    :param agent:
        The agent header to send in requests.

    :param normalize_unicode:
        Whether the JSON Serializer should normalize accented characters so that they
        can be encoded in Quickbase.
    """

    def __init__(
        self,
        table: Type[QuickbaseTable],
        user_token,
        agent="python",
        normalize_unicode=True,
        default_select=None,
    ):
        self.table = table
        self.serializer = RecordJsonSerializer(
            table_cls=self.table, normalize_unicode=normalize_unicode
        )
        self.api = QuickbaseApiClient(user_token, table.realm_hostname(), agent=agent)
        self._default_select = default_select

    @property
    def app_id(self):
        return self.table.app_id()

    @property
    def table_id(self):
        return self.table.__dbid__

    def get_app(self):
        """Get an app.

        https://developer.quickbase.com/operation/getApp
        """
        return self.api.get_app(self.app_id)

    def get_tables_for_app(self):
        """Get an tables for an app.

        https://developer.quickbase.com/operation/getAppTables
        """
        return self.api.get_tables_for_app(self.app_id)

    def get_table(self):
        """Get a table.

        https://developer.quickbase.com/operation/getTable
        """
        return self.api.get_table(self.app_id, self.table_id)

    def get_fields_for_table(self):
        """Get fields for a table.

        https://developer.quickbase.com/operation/getFields
        """
        return self.api.get_fields_for_table(self.table_id)

    @staticmethod
    def _get_field_id(field: Union[QuickbaseField, int]):
        return field if isinstance(field, int) else field.fid

    def get_field(self, field: Union[QuickbaseField, int]):
        """Get fields for a table.

        https://developer.quickbase.com/operation/getField

        :param field: either the field ID or a :class:`~QuickbaseField`
        """
        field_id = self._get_field_id(field)
        return self.api.get_field(field_id, self.table_id)

    def get_reports_for_table(self):
        """Get reports for a table.

        https://developer.quickbase.com/operation/getTableReports
        """
        return self.api.get_reports_for_table(self.table_id)

    def _get_report_id(self, report: Union[QuickbaseReport, str, int]):
        return (
            report
            if isinstance(report, int)
            else report.report_id
            if isinstance(report, QuickbaseReport)
            else self.table.get_report(report).report_id
        )

    def get_report(self, report):
        """Get report.

        https://developer.quickbase.com/operation/getRepor

        :param report: Either the report name to lookup, the report id, or a
            :class:`~QuickbaseReport` object.
        """
        report_id = self._get_report_id(report)
        return self.api.get_report(report_id, self.table_id)

    def run_report(self, report, skip=None, top=None):
        """Run report.

        https://developer.quickbase.com/operation/runReport.

        :param report:
            Either the report name to lookup, the report id, or a
            :class:`~QuickbaseReport` object.

        :param int skip:
            For paging (see Quickbase API)

        :param int top:
            For paging (see Quickbase API)
        """
        report_id = self._get_report_id(report)
        return self.api.run_report(report_id, self.table_id, skip=skip, top=top)

    def _encode_rec(self, rec):
        return (
            self.serializer.serialize(rec) if isinstance(rec, QuickbaseTable) else rec
        )

    def add_records(
        self,
        recs: List[Union[QuickbaseTable, Any]],
        merge_field_id=None,
        fields_to_return=None,
    ):
        """Add record.

        https://developer.quickbase.com/operation/upsert

        :param recs:
            A list of items that are either the raw record data to post, or the
            :class:`~QuickbaseTable` object/record.

        :param merge_field_id:
            The list of fields to merge on.

        :param fields_to_return:
            The list of field ID's to return (default None which means all).
        """
        data = [self._encode_rec(rec) for rec in recs]
        return self.api.add_records(
            table_id=self.table_id,
            data=data,
            merge_field_id=merge_field_id,
            fields_to_return=fields_to_return,
        )

    def add_record(self, rec, *args, **kwargs):
        """Aliased to :meth:`~add_records` making rec a list."""
        return self.add_records([rec], *args, **kwargs)

    def query(
        self, query_obj: QuickbaseQuery = None, raw=False, pager: ResponsePager = None
    ):
        """Do a query.

        https://developer.quickbase.com/operation/runQuery.

        See :mod:`~quickbase_client.query` for more.

        See :class:`~quickbase_client.ResponsePager` for handling pagination.

        If some fields are coming back as null, a common "gotcha" is that the
        Quickbase API by default only returns fields listed as "default" in a table.
        In that case you would have to explicitly specify a ``select`` in the
        query, or you can edit the fields in Quickbase to be default fields.

        :param query_obj:
            The :class:`~QuickbaseQuery` object to use. Note that this object also
            specifies the ``select``, ``group_by``, ``sort_by``, etc. So to specify
            those you need to specify them in the provided :class:`~QuickbaseQuery`.
            See its documentation for more details.

        :param raw:
            If true, returns a requests.Response, else the data is
            serialized to a table object.

        :param pager:
            A :class:`~ResponsePager` to handle making paginated requests.
        """
        pager = ResponsePager() if pager is None else pager
        query_obj = QuickbaseQuery(where=None) if query_obj is None else query_obj
        options = pager.get_options()
        options = {**options, **(query_obj.options if query_obj.options else {})}
        data = self.api.query(
            table_id=self.table_id,
            fields_to_select=query_obj.select,
            where_str=query_obj.where,
            sort_by=query_obj.sort_by,
            group_by=query_obj.group_by,
            options=options,
        )
        pager.update_from_metadata(data.json()["metadata"])
        return (
            data
            if raw
            else [self.serializer.deserialize(x) for x in data.json()["data"]]
        )

    def change_record_owner(self, rid, new_owner):
        """Use the legacy API to change a Record's owner.

        See https://help.quickbase.com/api-guide/change_record_owner.html

        :param rid:
            The record ID to change the owner of

        :param new_owner:
            The email address, or user ID, of the user to change to the owner.
        """
        return self.api.legacy_api.change_record_owner(self.table_id, rid, new_owner)


QuickBaseTableClient = QuickbaseTableClient  # alias - TODO - delete in future



from quickbase_client.client.api import QuickBaseApiClient
from quickbase_client.client.api import QuickbaseApiClient
from quickbase_client.client.legacy_client import QuickbaseLegacyApiClient
from quickbase_client.client.pager import ResponsePager
from quickbase_client.client.table_client import QuickBaseTableClient
from quickbase_client.client.table_client import QuickbaseTableClient
from quickbase_client.orm.app import QuickBaseApp
from quickbase_client.orm.app import QuickbaseApp
from quickbase_client.orm.field import QuickBaseField
from quickbase_client.orm.field import QuickBaseFieldType
from quickbase_client.orm.field import QuickbaseField
from quickbase_client.orm.field import QuickbaseFieldType
from quickbase_client.orm.report import QuickBaseReport
from quickbase_client.orm.report import QuickbaseReport
from quickbase_client.orm.table import QuickBaseTable
from quickbase_client.orm.table import QuickbaseTable
from quickbase_client.query.query_base import QuickBaseQuery
from quickbase_client.query.query_base import QuickbaseQuery

# NOTE - currently a bunch of duplicate aliases for QuickBase to Quickbase since this
# was originally released with everything prefixed as QuickBase. But since Quickbase
# is branding more to "Quickbase", this will eventually be the main naming for
# version 1.0 in an effort to keep more consistent.


__all__ = [
    "QuickBaseApiClient",
    "QuickbaseApiClient",
    "ResponsePager",
    "QuickBaseTableClient",
    "QuickbaseTableClient",
    "QuickbaseApp",
    "QuickBaseApp",
    "QuickBaseField",
    "QuickbaseField",
    "QuickbaseFieldType",
    "QuickBaseFieldType",
    "QuickBaseReport",
    "QuickbaseReport",
    "QuickBaseTable",
    "QuickbaseTable",
    "QuickBaseQuery",
    "QuickbaseQuery",
    "QuickbaseLegacyApiClient",
]



"""This module includes functions which create :class:`~QuickbaseQuery` objects.

These can be assembled in an AST-like fashion to build a complex query using
higher-level english-readable functions rather than going through the query
language (note you can always create a :class:`~QuickbaseQuery` and provide
the ``where`` string to use that).

Example:

.. code-block:: python

    schema = MyTable.schema
    my_query = and_(
        eq_(schema.date_opened, schema.date_created),
        on_or_before_(schema.date_closed, date(2020, 11, 16))
    )
    print(my_query.where) # ({'9'.EX.'_FID_1'}AND{'10'.OBF.'11-16-2020'})

All of the methods (except the two conjunction ones), take a
:class:`~QuickbaseField` and a value as a parameter. If you pass a `QuickbaseField`
for the value, it will compare to the actual field (see above). But note if you pass
an attribute of a QuickbaseTable class it would be the value in memory of that attribute.
If you want to compare to the actual field, use the schema property of the table or
:meth:`quickbase_client.QuickbaseTable.get_field_info`.


Note all of these methods are named with a trailing ``_`` to maintain consistency and
never clash with a python keyword or anything.
"""


from quickbase_client.query.query_base import QuickbaseQuery
from quickbase_client.query.query_utils import make_query_string


def qb_query_ast(func):
    def _wrap(*args, **kwargs):
        return QuickbaseQuery(where=func(*args, **kwargs))

    return _wrap


def _conjunction(kind, *clauses):
    return f"({kind.join([c.where for c in clauses])})"


# Note the decorator makes it tricky for sphinx to read the autodoc so docs
# are just in the docs folder for these functions.


@qb_query_ast
def or_(*clauses):
    return _conjunction("OR", *clauses)


@qb_query_ast
def and_(*clauses):
    return _conjunction("AND", *clauses)


@qb_query_ast
def contains_(field, val):
    return make_query_string(field, "CT", val)


@qb_query_ast
def not_contains_(field, val):
    return make_query_string(field, "XCT", val)


@qb_query_ast
def has_(field, val):
    """Has (HAS)."""
    return make_query_string(field, "HAS", val)


@qb_query_ast
def not_has_(field, val):
    """Not Has (XHAS)."""
    return make_query_string(field, "XHAS", val)


@qb_query_ast
def eq_(field, val):
    return make_query_string(field, "EX", val)


# QUESTION - don't understand this operator in their query language...
@qb_query_ast
def true_(field, val):
    return make_query_string(field, "TV", val)


@qb_query_ast
def not_eq_(field, val):
    return make_query_string(field, "XEX", val)


@qb_query_ast
def starts_with_(field, val):
    return make_query_string(field, "SW", val)


@qb_query_ast
def not_starts_width_(field, val):
    return make_query_string(field, "XSW", val)


@qb_query_ast
def before_(field, val):
    return make_query_string(field, "BF", val)


@qb_query_ast
def on_or_before_(field, val):
    return make_query_string(field, "OBF", val)


@qb_query_ast
def after_(field, val):
    return make_query_string(field, "AF", val)


@qb_query_ast
def on_or_after_(field, val):
    return make_query_string(field, "OAF", val)


@qb_query_ast
def during_(field, val):
    return make_query_string(field, "IR", val)


@qb_query_ast
def not_during_(field, val):
    return make_query_string(field, "XIR", val)


@qb_query_ast
def lt_(field, val):
    return make_query_string(field, "LT", val)


@qb_query_ast
def lte_(field, val):
    return make_query_string(field, "LTE", val)


@qb_query_ast
def gt_(field, val):
    return make_query_string(field, "GT", val)


@qb_query_ast
def gte_(field, val):
    return make_query_string(field, "GTE", val)



import attr


@attr.s(auto_attribs=True)
class QuickbaseReport(object):
    report_id: int
    name: str
    type: str = ""
    description: str = ""


QuickBaseReport = QuickbaseReport  # alias - TODO - delete in future



"""
The table module includes a class :class:`~QuickbaseTable` that facilitates an ORM
to QuickBase API JSON through a meta class also defined in this module.
"""

from typing import Dict

from quickbase_client.orm.app import QuickbaseApp
from quickbase_client.orm.field import QuickbaseField
from quickbase_client.orm.report import QuickbaseReport


# help from
# https://programmer.help/blogs/python-how-to-implement-orm-with-metaclasses.html


class QuickbaseTableMeta(type):
    """Meta-class which builds a dunder to store field data.

    Stores a mapping of attribute name --> QuickbaseFieldType in ``__mappings__``,

    """

    def __new__(mcs, name, bases, attrs):
        mappings = {}
        fidmap = {}
        _schema = QuickbaseTableSchema()

        for k, v in attrs.items():
            if isinstance(v, QuickbaseField):
                mappings[k] = v
                fidmap[v.fid] = k
                setattr(_schema, k, v)

        # Delete these properties that are already stored in the dictionary
        for k in mappings.keys():
            attrs.pop(k)

        attrs["__mappings__"] = mappings
        attrs["__fidmap__"] = fidmap
        attrs["__schema__"] = _schema
        return type.__new__(mcs, name, bases, attrs)

    @property
    def schema(cls):
        return cls.__schema__  # noqa


class QuickbaseTableSchema(object):
    """Used to get :class:`~QuickbaseField` data and not the value of the field."""

    pass


class QuickbaseTable(metaclass=QuickbaseTableMeta):
    """
    Base class for a table object.

    :ivar __dbid__: The string table ID (the part after /db in the URL).
    :ivar __tablename__: The english name of the table.
    :ivar __app__: The :class:`~QuickbaseApp` the table belongs to.
    :ivar __reports__: Lookup of :class:`~QuickbaseReport` objects for this table.
    :cvar schema: The object to reference field types (rather than field data).
    """

    __dbid__ = None
    __tablename__ = ""
    __app__: QuickbaseApp = None

    __reports__: Dict[str, QuickbaseReport] = {}

    def __init__(self, **kwargs):

        for attr, field_def in self.__mappings__.items():  # noqa
            v = kwargs.pop(attr) if attr in kwargs else None
            setattr(self, attr, v)

        for name, _ in kwargs.items():  # simple way to get the kwarg name if not empty
            raise AttributeError(f"no attribute named {name}")

    @classmethod
    def app_id(cls) -> str:
        """Alias to the app's ID."""
        return cls.__app__.app_id

    @classmethod
    def realm_hostname(cls) -> str:
        """Alias to the app's realm hostname."""
        return cls.__app__.realm_hostname

    @classmethod
    def get_field_info(cls, attr: str) -> QuickbaseField:
        """Get the field info for a given attribute rather than the data.

        :param attr: String name of the attribute.
        """
        return cls.__mappings__[attr]  # noqa

    @classmethod
    def get_report(cls, name: str):
        """Get a report by it's name.

        :param name: The name of the report
        """
        return cls.__reports__[name]

    @classmethod
    def get_attr_from_fid(cls, fid: int):
        """Lookup an attribute name by it's field ID.

        :param fid: The field ID.
        """
        return cls.__fidmap__[fid]  # noqa

    @classmethod
    def client(cls, user_token: str):
        """Factory method to create a :class:`~QuickbaseTableClient` for this table.

        :param user_token: The user token for authentication.
        """
        from quickbase_client.client.table_client import QuickbaseTableClient

        return QuickbaseTableClient(cls, user_token)


QuickBaseTable = QuickbaseTable  # alias - TODO - delete in future.



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['quickbase_client',
 'quickbase_client.client',
 'quickbase_client.orm',
 'quickbase_client.query',
 'quickbase_client.tools',
 'quickbase_client.utils']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.3.0,<21.0.0', 'requests>=2.24.0,<3.0.0', 'stringcase>=1.2.0,<2.0.0']

entry_points = \
{'console_scripts': ['qbc = quickbase_client.tools.qbc:main']}

setup_kwargs = {
    'name': 'quickbase-client',
    'version': '0.5.2',
    'description': 'A Quickbase Python API Client Generator',
    'long_description': '#####################\nQuickbase-Client\n#####################\n\nA High-Level Quickbase Python API Client & Model Generator\n\n\n.. image:: https://gitlab.com/tkutcher/quickbase-client/badges/dev/pipeline.svg\n    :target: https://gitlab.com/tkutcher/quickbase-client/-/commits/dev\n    :alt: Pipeline Status\n\n.. image:: https://gitlab.com/tkutcher/quickbase-client/badges/dev/coverage.svg\n    :target: https://gitlab.com/tkutcher/quickbase-client/-/commits/dev\n    :alt: Coverage Report\n\n.. image:: https://readthedocs.org/projects/quickbase-client/badge/?version=latest\n    :target: https://quickbase-client.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. image:: https://badge.fury.io/py/quickbase-client.svg\n    :target: https://badge.fury.io/py/quickbase-client\n    :alt: PyPI\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: Black Code Style\n\n|\n\n\n*Quickbase-Client is a library for interacting with Quickbase applications through their\nRESTful JSON API (https://developer.quickbase.com/). It has features to generate model classes\nfor tables in your Quickbase app, and provides high level classes to interface between Python\nobjects and the Quickbase tables.*\n\n|\n\n\n.. inclusion-marker-do-not-remove\n\nQuick Start\n============\n\n\nInstallation\n____________\n\nInstallation can be done through pip:\n\n.. code-block:: bash\n\n    pip install quickbase-client\n\nThis will install both the library ``quickbase_client``, and a command line tool ``qbc`` for\nrunning some handy scripts.\n\n\nGenerating your Models\n----------------------\n\nTo interact and authenticate with your Quickbase applications you need a User Token. You can read\nthe Quickbase documentation `here <https://developer.quickbase.com/auth>`_ on how to create one.\nIt is recommended to set an environment variable ``QB_USER_TOKEN`` with this value:\n\n.. code-block:: bash\n\n    export QB_USER_TOKEN=mytokenfromquickbase;\n\n\nNext, say you have a hypothetical Quickbase Application named MyApp at\n``https://foo.quickbase.com/db/abcdef`` that has tables for tracking things\nagainst a repository like Issues & Pipelines.\n\n\n.. image:: /images/example_table.png\n    :width: 500\n    :alt: Example Table\n\n|\n\nRunning the following:\n\n.. code-block:: bash\n\n    qbc run model-generate -a https://foo.quickbase.com/db/abcdef\n\nWould generate a directory structure like\n\n::\n\n    models\n    ├── __init__.py\n    └── my_app\n        ├── __init__.py\n     \xa0\xa0 ├── app.py\n    \xa0\xa0  ├── github_issue.py\n    \xa0\xa0  └── gitlab_pipeline.py\n\nAnd classes like ``GitHubIssue`` where you can interact with the data model through a Python object.\n\n\nWriting Records to Quickbase\n----------------------------\n\nClasses like ``GitHubIssue`` that subclass ``QuickbaseTable`` also get a factory class-method\n``client(user_tok)`` which creates an instance of the higher-level ``QuickbaseTableClient`` to\nmake API requests for things related to that table:\n\n.. code-block:: python\n\n    client = GitHubIssue.client(user_tok=os.environ[\'QB_USER_TOKEN\'])\n    new_issue = GitHubIssue(\n        title=\'Something broke\',   # you get friendly-kwargs for fields without worrying about ID\'s\n        description=\'Please fix!\',\n        date_opened=date.today()   # things like Python date objects will be serialized\n    )\n    response = client.add_record(new_issue)\n    print(response.json())  # all methods (except for query) return the requests Response object\n\n\nQuerying Records from Quickbase\n-------------------------------\n\nYou can also use the client object to send queries to the Quickbase API through the ``query``\nmethod. This method will serialize the data back in to a Python object. The `query` method on the\ntable class takes a ``QuickbaseQuery`` object which is high level wrapper around the parameters\nneeded to make a query.\n\nNotably, the ``where`` parameter for specifying the query string. There is one (and in the future\nthere will be more) implementation of this which allows you to build query-strings through\nhigher-level python functions.\n\nYou can use the methods exposed in the ``quickbase_client.query`` module like so:\n\n.. code-block:: python\n\n    # convention to append an underscore to these methods to avoid clashing\n    # with any python keywords\n    from quickbase_client.query import on_or_before_\n    from quickbase_client.query import eq_\n    from quickbase_client.query import and_\n\n    schema = GitHubIssue.schema\n    q = and_(\n        eq_(schema.date_opened, schema.date_created),\n        on_or_before_(schema.date_closed, date(2020, 11, 16))\n    )\n    print(q.where)  # ({\'9\'.EX.\'_FID_1\'}AND{\'10\'.OBF.\'11-16-2020\'})\n    recs = client.query(q)  # recs will be GitHubIssue objects unless passing raw=True\n    print([str(r) for r in recs])  # [\'<GitHubIssue title="Made And Closed Today" id="10000">\']\n\n\n\nControlling Lower-Level API Calls\n---------------------------------\n\nLastly, say you want to deal with just posting the specific json/data Quickbase is looking for.\nThe ``QuickbaseTableClient`` object wraps the lower-level ``QuickbaseApiClient`` object which has\nmethods for just sending the actual data (with an even lower-level utility\n``QuickbaseRequestFactory`` you could also use). These classes manage hanging on to the user token,\nand the realm hostname, etc. for each request that is made.\n\nFor example, note the signature of ``query`` in ``QuickbaseApiClient``:\n\n.. code-block:: python\n\n    def query(self, table_id, fields_to_select=None, where_str=None,\n              sort_by=None, group_by=None, options=None):\n\n\nYou can get to this class by going through the table client: ``api = client.api``, or from\ninstantiating it directly ``api = QuickbaseApiClient(my_user_token, my_realm)``\n\nWith this, we could make the exact same request as before:\n\n.. code-block:: python\n\n    api = QuickbaseApiClient(user_token=\'my_token\', realm_hostname=\'foo.quickbase.com\')\n    response = api.query(\n        table_id=\'abcdef\',\n        where_str="({\'9\'.EX.\'_FID_1\'}AND{\'10\'.OBF.\'11-16-2020\'})")\n    data = response.json()\n\n\n.. exclusion-marker-do-not-remove\n\nMore Resources\n==============\n- `examples </examples>`_ directory.\n- `CONTRIBUTING </CONTRIBUTING.md>`_\n- `LICENSE </LICENSE.md>`_\n\n\nOther Notes\n====================\n\n\nCurrently a bunch of duplicate aliases for ``QuickBase`` to ``Quickbase`` since this\nwas originally released with everything prefixed as ``QuickBase-``. But since Quickbase\nis branding more to "Quickbase", this will eventually be the main naming for\nversion 1.0 in an effort to keep more consistent. So prefer to use `Quickbase-` prefixed classes\nas in the future the other aliases will be dropped.\n',
    'author': 'Tim Kutcher',
    'author_email': 'tim@tkutcher.com',
    'maintainer': 'Tim Kutcher',
    'maintainer_email': 'tim@tkutcher.com',
    'url': 'https://github.com/tkutcher/quickbase-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
import sys
from typing import Optional

from quickbase_client.client.api import QuickbaseApiClient
from quickbase_client.orm.field import get_field_type_by_string
from quickbase_client.tools.script import Script
from quickbase_client.utils.pywriting_utils import BasicPyFileWriter
from quickbase_client.utils.pywriting_utils import PyFileWriter
from quickbase_client.utils.pywriting_utils import PyPackageWriter
from quickbase_client.utils.string_utils import id_from_iso_string
from quickbase_client.utils.string_utils import make_unique_var_name
from quickbase_client.utils.string_utils import make_var_name
from quickbase_client.utils.string_utils import parse_realm_and_app_id_from_url


class AppPyFileWriter(PyFileWriter):
    def __init__(self):
        self.pyfile = BasicPyFileWriter()

    def add_header_comments_and_imports(self, app_name, description):
        self.pyfile.add_comment(
            f"app.py - model QuickbaseApp for {app_name} QuickBase App."
        ).add_comment(f"{app_name} - {description}").space().add_line(
            "from quickbase_client.orm.app import QuickbaseApp"
        ).space()

    def add_module_vars(self, version_id):
        self.pyfile.add_line(f"__versionid__ = '{version_id}'").space()

    def add_app_var(self, app_id, app_name, var_name, realm_hostname):
        self.pyfile.add_line(
            f"{var_name} = QuickbaseApp(app_id='{app_id}', name='{app_name}', "
            f"realm_hostname='{realm_hostname}')"
        ).space()

    def get_file_as_string(self):
        return self.pyfile.get_file_as_string()


class TablePyFileWriter(PyFileWriter):
    def __init__(self):
        self.pyfile = BasicPyFileWriter()
        self.field_vars = {}
        self.formula_outlet: Optional[BasicPyFileWriter] = None

    def add_header_comments_and_imports(
        self, app_import_path, app_var_name, file_name, table_name
    ):
        self.pyfile.add_comment(
            f"{file_name}.py - QuickbaseTable for {table_name}"
        ).space().add_line("from quickbase_client import QuickbaseField").add_line(
            "from quickbase_client import QuickbaseFieldType as Qb"
        ).add_line(
            "# from quickbase_client import QuickbaseReport"
        ).add_line(
            "from quickbase_client import QuickbaseTable"
        ).space().add_line(
            f"from {app_import_path}.app import {app_var_name}"
        ).space().space()
        self.formula_outlet = self.pyfile.make_ref()

    def add_table_class_decl(self, table_ident, table_id, app_var_name):
        class_name = make_var_name(table_ident, case="pascal")
        self.pyfile.space().add_line(
            f"class {class_name}(QuickbaseTable):"
        ).indent().add_line(f"__dbid__ = '{table_id}'").add_line(
            f"__app__ = {app_var_name}"
        ).space().add_line(
            "date_created = QuickbaseField(fid=1, field_type=Qb.DATETIME)"
        ).add_line(
            "date_modified = QuickbaseField(fid=2, field_type=Qb.DATETIME)"
        ).add_line(
            "recordid = QuickbaseField(fid=3, field_type=Qb.NUMERIC)"
        ).add_line(
            "record_owner = QuickbaseField(fid=4, field_type=Qb.USER)"
        ).add_line(
            "last_modified = QuickbaseField(fid=5, field_type=Qb.USER)"
        ).space()

    def add_table_field(self, field_name, field_id, field_kind, properties):
        if field_id < 6:
            return  # the special reserved ones already accounted for more or less.
        var_name = make_unique_var_name(field_name, taken=self.field_vars)
        field_kind = str(get_field_type_by_string(field_kind)).split(".")[1]

        formula_str = ""
        if properties["formula"] != "":
            self.formula_outlet.add_line(
                f"{var_name}_formula = '''{properties['formula']}'''"
            ).space()
            formula_str = f", formula={var_name}_formula"

        self.pyfile.add_line(
            f"{var_name} = QuickbaseField("
            f"fid={field_id}, "
            f"field_type=Qb.{field_kind}{formula_str})"
        )

    def done_writing(self):
        self.formula_outlet.space()

    def get_file_as_string(self):
        return self.pyfile.get_file_as_string()


class ModelGenerator(Script):
    registration_name = "model-generate"

    def __init__(
        self, realm_hostname, app_id, user_token, pkg_dir=None, table_ids=None
    ):
        self.realm_hostname = realm_hostname
        self.app_id = app_id
        self.user_token = user_token
        self.app_var_name = None
        self.table_vars = {}
        self.pkg_import_stmt = ""
        self.table_ids = table_ids if table_ids else []

        if pkg_dir is None:
            models_pkg = PyPackageWriter(pkg_name="models", parent_dir=os.getcwd())
            if not os.path.exists(models_pkg.pkg_path):
                models_pkg.write()
            pkg_dir = models_pkg.pkg_path
            self.pkg_import_stmt = "models"

        self.pkg_writer = PyPackageWriter(pkg_name="PLACEHOLDER", parent_dir=pkg_dir)

    @staticmethod
    def add_argparse_args(parser):
        parser.add_argument(
            "-a",
            "--app-url",
            nargs="?",
            required=True,
            help="the URL of the home page of the app",
        )
        parser.add_argument(
            "-d",
            "--pkg-dir",
            nargs="?",
            default=None,
            help='the directory to put the package in, defaults to "models"',
        )
        parser.add_argument(
            "-t",
            "--user-tok",
            nargs="?",
            default=None,
            help="the user token to authenticate - if not provided will read from "
            "environment variable QB_USER_TOKEN",
        )
        parser.add_argument(
            "-i",
            "--include",
            nargs="?",
            default=None,
            action="append",
            help="ID or name of a table to include - can be specified "
            "multiple times; if present, excludes all other tables",
        )

    @staticmethod
    def instantiate_from_ns(ns) -> "Script":
        realm_hostname, app_id = parse_realm_and_app_id_from_url(ns.app_url)
        try:
            user_tok = (
                ns.user_tok if ns.user_tok is not None else os.environ["QB_USER_TOKEN"]
            )
            return ModelGenerator(
                realm_hostname,
                app_id,
                user_token=user_tok,
                pkg_dir=ns.pkg_dir,
                table_ids=ns.include,
            )
        except (KeyError, OSError):
            raise EnvironmentError(
                "ERROR: expected either a -t argument or a "
                "QB_USER_TOKEN environment variable"
            )

    def add_app_file(self, app_data):
        name = app_data["name"]
        self.pkg_writer.pkg_name = make_var_name(name, case="snake")
        self.pkg_import_stmt += (
            "." if self.pkg_import_stmt else ""
        ) + self.pkg_writer.pkg_name
        description = app_data["description"]
        updated = app_data["updated"]
        versionid = id_from_iso_string(updated)
        self.app_var_name = make_var_name(name, case="pascal")
        w = AppPyFileWriter()
        w.add_header_comments_and_imports(app_name=name, description=description)
        w.add_module_vars(versionid)
        w.add_app_var(self.app_id, name, self.app_var_name, self.realm_hostname)
        self.pkg_writer.add_module("app", w)

    def add_table_file(self, table, fields):
        table_ident = table["singleRecordName"]
        file_name = make_var_name(table_ident)
        if file_name in self.pkg_writer.modules:
            table_ident = table["alias"].replace("_DBID_", "")
            file_name = make_var_name(table_ident)
        w = TablePyFileWriter()
        w.add_header_comments_and_imports(
            self.pkg_import_stmt, self.app_var_name, file_name, table["name"]
        )
        w.add_table_class_decl(table_ident, table["id"], self.app_var_name)
        for f in sorted(fields, key=lambda f: int(f["id"])):
            w.add_table_field(f["label"], f["id"], f["fieldType"], f["properties"])
        w.done_writing()
        self.pkg_writer.add_module(file_name, w)

    def run(self):
        api = QuickbaseApiClient(
            user_token=self.user_token, realm_hostname=self.realm_hostname
        )

        app_data = api.get_app(self.app_id)
        if not app_data.ok:
            if "invalid dbid" in app_data.json()["message"].lower():
                sys.stderr.write(
                    "URL should be to the app home page, not a specific table.\n"
                )
                return False
            else:
                sys.stderr.write(app_data.text + "\n")
                raise ValueError(
                    f"received {app_data.status_code} from QuickBase API -"
                    f"see error log above"
                )
        self.add_app_file(app_data.json())

        tables = api.get_tables_for_app(self.app_id)
        for table in tables.json():
            table_id = table["id"]
            table_ident = table["singleRecordName"]
            table_var = make_var_name(table_ident)
            if not len(self.table_ids) or any(
                x in self.table_ids for x in [table_var, table_ident, table_id]
            ):
                # if not table_ids have been specified, or they have but this table's ID,
                # single record name, or var name are in the list, generate it
                fields = api.get_fields_for_table(table_id)
                self.add_table_file(table, fields.json())

        self.pkg_writer.write()
        return True



import abc
from typing import Type

from quickbase_client.tools.model_generate import ModelGenerator
from quickbase_client.tools.script import Script


class ScriptManager(object):
    def __init__(self):
        self.scripts = {}

    def all_scripts(self):
        return self.scripts.keys()

    def register_script(self, script_cls: Type[Script]):
        name = script_cls.registration_name
        if name in self.scripts:
            raise KeyError(f"Script {name} already registered!")
        self.scripts[name] = script_cls

    def get_script_by_name(self, name):
        return self.scripts[name]


class ScriptLoader(abc.ABC):
    @abc.abstractmethod
    def load_scripts(self) -> ScriptManager:
        pass


class CoreScriptLoader(ScriptLoader):
    def load_scripts(self):
        mgr = ScriptManager()
        mgr.register_script(ModelGenerator)
        return mgr



import logging
import threading
from datetime import datetime

from quickbase_client import QuickbaseTableClient

_mock_record = logging.LogRecord(
    name="", level=0, pathname="", lineno=1, msg="", args=None, exc_info=None
)


class QuickbaseLogHandler(logging.Handler):
    """Class for sending logs to a specific Quickbase table.

    You supply it with a table client to the logs table, and the handler will send logs
    to that table in a non-blocking background thread.

    This uses the higher-level QuickbaseTableClient APIs. So you will have to
    create a class for your table you want to send logs to.
    """

    def __init__(self, logs_table_client: QuickbaseTableClient):
        self.client = logs_table_client

        try:
            self.record_factory(_mock_record)
        except Exception:
            raise TypeError(
                "Invalid record_factory - could not create a log record. If using"
                'the default, either ensure your table has properties "when", '
                '"level", and "message", OR implement a custom record_factory.'
            )

        super().__init__()

    @staticmethod
    def with_record_factory(logs_table_client: QuickbaseTableClient, record_factory):
        """Create a logger using a function to make records.

        :param logs_table_client:
            The QuickbaseTableClient for the logs to go to.

        :param record_factory:
            A function which takes a `logging.LogRecord` and creates the relevant
            record (instance of a :class:`~QuickbaseTable`).
        """

        class _CustomRecordFactoryHandler(QuickbaseLogHandler):
            def record_factory(self, record: logging.LogRecord):
                return record_factory(record)

        return _CustomRecordFactoryHandler(logs_table_client)

    def record_factory(self, record: logging.LogRecord):
        """
        Create a :class:`~QuickbaseTable` record object given a LogRecord. By default,
        this assumes the associated table, under the handlers table client, has
        properties ``when``, ``level``, and ``message``.

        :param record: The logging.LogRecord.
        :return: A QuickbaseTable record object.
        """
        return self.client.table(
            when=datetime.utcnow(),
            level=record.levelno,
            message=record.msg,
        )

    def emit(self, record):
        """Calls :meth:`~record_factory` and starts a thread to send it to Quickbase."""
        t = threading.Thread(target=self._do_emit, args=[record])
        t.start()

    def _do_emit(self, record):
        qb_rec = self.record_factory(record)
        return self.client.add_record(qb_rec)



import abc
import json
from datetime import date
from datetime import datetime
from typing import Dict
from typing import Type

from quickbase_client.orm.field import QuickbaseFieldType
from quickbase_client.orm.table import QuickbaseTable
from quickbase_client.utils.string_utils import normalize_unicode


# TODO - encoder should be build dynamically from app date props.


class QuickbaseJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat().split(".")[0]
        elif isinstance(o, date):
            return o.isoformat()
        return super().default(o)  # pragma: no cover


class RecordSerializer(abc.ABC):
    @abc.abstractmethod
    def serialize(self, record: "QuickbaseTable"):
        pass  # pragma: no cover

    @abc.abstractmethod
    def deserialize(self, data):
        pass  # pragma: no cover


class RecordJsonSerializer(RecordSerializer):
    def __init__(self, table_cls: Type[QuickbaseTable], normalize_unicode: bool = True):
        self.table_cls = table_cls
        self.normalize_unicode = normalize_unicode

    def serialize(self, record: "QuickbaseTable") -> Dict:
        o = {}
        for attr, v in record.__dict__.items():
            if attr[0] == "_" or v is None:
                continue
            field_info = record.get_field_info(attr)
            if field_info.field_type == QuickbaseFieldType.DATE and isinstance(
                v, datetime
            ):
                v = v.date()
            if (
                self.normalize_unicode
                and field_info.field_type
                in [
                    QuickbaseFieldType.TEXT,
                    QuickbaseFieldType.TEXT_MULTILINE,
                    QuickbaseFieldType.TEXT_MULTI_SELECT,
                    QuickbaseFieldType.TEXT_MULTIPLE_CHOICE,
                ]
                and isinstance(v, str)
            ):
                v = normalize_unicode(v)
            o[field_info.fid] = {"value": v}
        return o

    def deserialize(self, data):
        c = self.table_cls()
        for fid, v in data.items():
            a = c.get_attr_from_fid(int(fid))
            field_info = c.get_field_info(a)

            # could be cleaner elsewhere
            try:
                if field_info.field_type == QuickbaseFieldType.DATE:
                    val = datetime.strptime(v["value"], "%Y-%m-%d").date()
                elif field_info.field_type == QuickbaseFieldType.DATETIME:
                    val = datetime.fromisoformat(v["value"].rstrip("Z"))
                else:
                    val = v["value"]
            except (ValueError, KeyError):
                val = None
            setattr(c, a, val)

        return c



import abc
import os
from pathlib import Path
from typing import Dict
from typing import TextIO


class PyPackageWriter(object):
    def __init__(self, pkg_name, parent_dir=None):
        self.pkg_name = pkg_name
        self.parent_dir = os.getcwd() if parent_dir is None else parent_dir
        self.modules: Dict[str, PyFileWriter] = {}
        self.init_module = BasicPyFileWriter()
        self.add_module("__init__", self.init_module)

    @property
    def pkg_path(self):
        return os.path.join(self.parent_dir, self.pkg_name)

    def has_module_name(self, module_name):
        return module_name in self.modules

    def add_module(self, module_name, pyfile: "PyFileWriter"):
        self.modules[module_name] = pyfile

    def write(self):
        for module_name, pyfile in self.modules.items():
            Path(self.pkg_path).mkdir(parents=True, exist_ok=True)
            with open(os.path.join(self.pkg_path, f"{module_name}.py"), "w") as f:
                pyfile.dump(f)


class PyFileWriter(abc.ABC):
    @abc.abstractmethod
    def get_file_as_string(self):
        pass

    def dump(self, dest: TextIO):
        dest.write(self.get_file_as_string())


class BasicPyFileWriter(PyFileWriter):
    """
    A simple class to help writing out python files. Could alternatively render jinja templates
    but this gives a little more potential flexibility and the general files we are trying
    to write are not complex enough to really need a jinja template. Plus this is just more fun.

    This lets you call make_ref and get back another writer object that will be expanded in that
    spot. I.e. so you can insert code earlier in the file at a later point in the algorithm.
    """

    # @attr.s(auto_attribs=True)
    # class Line:
    #     lineno: int
    #     s: str

    def __init__(self, indent="    "):
        super().__init__()
        self._indent = indent
        self.lines = []
        self.level = 0

    def indent(self):
        self.level += 1
        return self

    def dedent(self):
        self.level -= 1
        return self

    def make_ref(self):
        w = BasicPyFileWriter(indent=self._indent)
        self.lines.append(w)
        return w

    def add_line(self, s, level=None):
        level = self.level if level is None else level
        self.lines.append(f"{self._indent * level}{s}")
        return self

    def space(self):
        return self.add_line("")

    def add_comment(self, comment, level=0):
        return self.add_line(f"# {comment}", level)

    def get_file_as_string(self):
        lns = [
            ln.get_file_as_string() if isinstance(ln, PyFileWriter) else ln
            for ln in self.lines
        ]
        return "\n".join(lns)



import attr


@attr.s(auto_attribs=True)
class QuickbaseApp(object):
    """Class for a QuickBase app."""

    app_id: str
    realm_hostname: str
    name: str


QuickBaseApp = QuickbaseApp  # alias - TODO - delete in future



import argparse
import sys

from quickbase_client.tools.script_loader import CoreScriptLoader

scripts = CoreScriptLoader().load_scripts()


def _run(ns):
    print("running...")
    script_instance = ns.script_cls.instantiate_from_ns(ns)
    if script_instance:
        retval = script_instance.run()
        if retval:
            print("done!")
        return retval
    else:
        return 1


def main(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("--show-stacktrace", action="store_true")

    cmd_parsers = parser.add_subparsers(help="command help")

    run_parser = cmd_parsers.add_parser("run", help="run a script.")
    run_parser.set_defaults(func=_run)

    script_parsers = run_parser.add_subparsers(help="script help")
    for script_name in scripts.all_scripts():
        script = scripts.get_script_by_name(script_name)
        script_parser = script_parsers.add_parser(script_name)
        script.add_argparse_args(script_parser)
        script_parser.set_defaults(script_cls=script)

    ns = parser.parse_args() if args is None else parser.parse_args(args)

    try:
        return ns.func(ns)
    except Exception as e:  # noqa
        if ns.show_stacktrace:
            raise e
        sys.stderr.write(str(e) + "\n"),
        sys.stderr.write("\trerun with qbc --show-stacktrace for more output\n")
        sys.stderr.write("\texample: qbc --show-stacktrace run blah\n")
        exit(1)


if __name__ == "__main__":
    main()



from datetime import date
from datetime import datetime

from quickbase_client.orm.field import QuickbaseField

DATE_FMT = "%m-%d-%Y"
DATETIME_FMT = "%m-%d-%Y %l:%M%p"


def query_value_stringify(val, _quote=True):
    def quoter(v):
        return f"'{v}'" if _quote else v

    if isinstance(val, datetime):
        return quoter(val.strftime(DATETIME_FMT))
    if isinstance(val, date):
        return quoter(val.strftime(DATE_FMT))
    if isinstance(val, bool):
        return quoter("true" if val else "false")
    if isinstance(val, list):
        return quoter("; ".join([query_value_stringify(v, _quote=False) for v in val]))
    if isinstance(val, QuickbaseField):
        return quoter(f"_FID_{val.fid}")
    if isinstance(val, int) or isinstance(val, float):
        return str(val)  # don't quote
    return quoter(str(val))


def make_query_string(field, operator, matching_value):
    fid = field.fid if isinstance(field, QuickbaseField) else field
    val = query_value_stringify(matching_value)
    return f"{{'{fid}'.{operator}.{val}}}"
